@@ -1,9 +1,9 @@
-# -*- coding: ISO-8859-1 -*-
+# -*- encoding: utf-8 -*-
 #
 #
-# Copyright (C) 2002-2004 Jrg Lehmann <joergl@users.sourceforge.net>
+# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
 # Copyright (C) 2003-2004 Michael Schindler <m-schindler@users.sourceforge.net>
-# Copyright (C) 2002-2006 Andr Wobst <wobsta@users.sourceforge.net>
+# Copyright (C) 2002-2011 André Wobst <wobsta@users.sourceforge.net>
 #
 # This file is part of PyX (http://pyx.sourceforge.net/).
 #
@@ -22,8 +22,6 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 
 
-from __future__ import nested_scopes
-
 import math, re, string, warnings
 from pyx import canvas, path, trafo, unit
 from pyx.graph import style
@@ -133,6 +131,7 @@ class graph(canvas.canvas):
         canvas.canvas.__init__(self)
         self.axes = {}
         self.plotitems = []
+        self.keyitems = []
         self._calls = {}
         self.didranges = 0
         self.didstyles = 0
@@ -250,15 +249,26 @@ class graph(canvas.canvas):
 
         self.didstyles = 1
 
-    def doplot(self, plotitem):
-        if self.did(self.doplot, plotitem):
+    def doplotitem(self, plotitem):
+        if self.did(self.doplotitem, plotitem):
             return
         self.dostyles()
         plotitem.draw(self)
 
-    def dodata(self):
+    def doplot(self):
         for plotitem in self.plotitems:
-            self.doplot(plotitem)
+            self.doplotitem(plotitem)
+
+    def dodata(self):
+        warnings.warn("dodata() has been deprecated. Use doplot() instead.")
+        self.doplot()
+
+    def dokeyitem(self, plotitem):
+        if self.did(self.dokeyitem, plotitem):
+            return
+        self.dostyles()
+        if plotitem.title is not None:
+            self.keyitems.append(plotitem)
 
     def dokey(self):
         raise NotImplementedError
@@ -266,7 +276,7 @@ class graph(canvas.canvas):
     def finish(self):
         self.dobackground()
         self.doaxes()
-        self.dodata()
+        self.doplot()
         self.dokey()
 
 
@@ -497,9 +507,10 @@ class graphxy(graph):
         if self.did(self.dokey):
             return
         self.dobackground()
-        self.dostyles()
+        for plotitem in self.plotitems:
+            self.dokeyitem(plotitem)
         if self.key is not None:
-            c = self.key.paint(self.plotitems)
+            c = self.key.paint(self.keyitems)
             bbox = c.bbox()
             def parentchildalign(pmin, pmax, cmin, cmax, pos, dist, inside):
                 ppos = pmin+0.5*(cmax-cmin)+dist+pos*(pmax-pmin-cmax+cmin-2*dist)
@@ -631,7 +642,7 @@ class graphxyz(graphxy):
                 else:
                     self.axes[axisname] = axis.linkedaxis(self.axes[okey], axisname)
         if not axes.has_key("z"):
-            self.axes["z"] = axis.anchoredaxis(axis.linear(), self.texrunner, axisname)
+            self.axes["z"] = axis.anchoredaxis(axis.linear(), self.texrunner, "z")
 
         if self.axes.has_key("x"):
             self.xbasepath = self.axes["x"].basepath
@@ -686,7 +697,7 @@ class graphxyz(graphxy):
             yaxis = self.axes["y"]
         if zaxis is None:
             zaxis = self.axes["z"]
-        return self.vpos_pt(xaxis.convert(x), yaxis.convert(y), zaxis.convert(y))
+        return self.vpos_pt(xaxis.convert(x), yaxis.convert(y), zaxis.convert(z))
 
     def pos(self, x, y, z, xaxis=None, yaxis=None, zaxis=None):
         if xaxis is None:
@@ -695,7 +706,7 @@ class graphxyz(graphxy):
             yaxis = self.axes["y"]
         if zaxis is None:
             zaxis = self.axes["z"]
-        return self.vpos(xaxis.convert(x), yaxis.convert(y), zaxis.convert(y))
+        return self.vpos(xaxis.convert(x), yaxis.convert(y), zaxis.convert(z))
 
     def vpos_pt(self, vx, vy, vz):
         x, y = self.projector.point(2*self.xscale*(vx - 0.5),
@@ -731,7 +742,7 @@ class graphxyz(graphxy):
 
     def vgeodesic_el(self, vx1, vy1, vz1, vx2, vy2, vz2):
         """returns a geodesic path element between two points in graph coordinates"""
-        return path.lineto_pt(*(self.vpos_pt(vx1, vy1, vz1) + self.vpos_pt(vx2, vy2, vz2)))
+        return path.lineto_pt(*self.vpos_pt(vx2, vy2, vz2))
 
     def vcap_pt(self, coordinate, length_pt, vx, vy, vz):
         """returns an error cap path for a given coordinate, lengths and
@@ -847,7 +858,7 @@ class graphxyz(graphxy):
                                                                        lambda vz: self.vtickdirection(0, 1, vz, 1, 0, vz),
                                                                        self.zvgridpath))
         elif axisname == "z4":
-            self.axes["z4"].setpositioner(positioner.flexlineaxispos_pt(lambda vz: self.vpos_pt(0, 0, vz),
+            self.axes["z4"].setpositioner(positioner.flexlineaxispos_pt(lambda vz: self.vpos_pt(1, 1, vz),
                                                                        lambda vz: self.vtickdirection(1, 1, vz, 0, 0, vz),
                                                                        self.zvgridpath))
         else:
@@ -875,9 +886,10 @@ class graphxyz(graphxy):
         if self.did(self.dokey):
             return
         self.dobackground()
-        self.dostyles()
+        for plotitem in self.plotitems:
+            self.dokeyitem(plotitem)
         if self.key is not None:
-            c = self.key.paint(self.plotitems)
+            c = self.key.paint(self.keyitems)
             bbox = c.bbox()
             def parentchildalign(pmin, pmax, cmin, cmax, pos, dist, inside):
                 ppos = pmin+0.5*(cmax-cmin)+dist+pos*(pmax-pmin-cmax+cmin-2*dist)
