@@ -1,9 +1,9 @@
-# -*- coding: ISO-8859-1 -*-
+# -*- encoding: utf-8 -*-
 #
 #
-# Copyright (C) 2002-2006 Jrg Lehmann <joergl@users.sourceforge.net>
+# Copyright (C) 2002-2006 Jörg Lehmann <joergl@users.sourceforge.net>
 # Copyright (C) 2003-2005 Michael Schindler <m-schindler@users.sourceforge.net>
-# Copyright (C) 2002-2006 Andr Wobst <wobsta@users.sourceforge.net>
+# Copyright (C) 2002-2011 André Wobst <wobsta@users.sourceforge.net>
 #
 # This file is part of PyX (http://pyx.sourceforge.net/).
 #
@@ -21,17 +21,8 @@
 # along with PyX; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 
-from __future__ import nested_scopes
-
 import math
-from math import cos, sin, tan, acos, pi
-try:
-    from math import radians, degrees
-except ImportError:
-    # fallback implementation for Python 2.1
-    def radians(x): return x*pi/180
-    def degrees(x): return x*180/pi
-
+from math import cos, sin, tan, acos, pi, radians, degrees
 import trafo, unit
 from normpath import NormpathException, normpath, normsubpath, normline_pt, normcurve_pt
 import bbox as bboxmodule
@@ -41,20 +32,6 @@ from normpath import set
 # normpath's invalid is available as an external interface
 from normpath import invalid
 
-try:
-    sum([])
-except NameError:
-    # fallback implementation for Python 2.2 and below
-    def sum(list):
-        return reduce(lambda x, y: x+y, list, 0)
-
-try:
-    enumerate([])
-except NameError:
-    # fallback implementation for Python 2.2 and below
-    def enumerate(list):
-        return zip(xrange(len(list)), list)
-
 # use new style classes when possible
 __metaclass__ = type
 
@@ -327,6 +304,12 @@ class closepath(pathitem):
         file.write("closepath\n")
 
 
+class pdfmoveto_pt(normline_pt):
+
+    def outputPDF(self, file, writer):
+        pass
+
+
 class moveto_pt(pathitem):
 
     """Start a new subpath and set current point to (x_pt, y_pt) (coordinates in pts)"""
@@ -349,6 +332,9 @@ class moveto_pt(pathitem):
     def createnormpath(self, epsilon=_marker):
         if epsilon is _marker:
             return normpath([normsubpath([normline_pt(self.x_pt, self.y_pt, self.x_pt, self.y_pt)])])
+        elif epsilon is None:
+            return normpath([normsubpath([pdfmoveto_pt(self.x_pt, self.y_pt, self.x_pt, self.y_pt)],
+                                         epsilon=epsilon)])
         else:
             return normpath([normsubpath([normline_pt(self.x_pt, self.y_pt, self.x_pt, self.y_pt)],
                                          epsilon=epsilon)])
