@@ -1,9 +1,9 @@
-# -*- coding: ISO-8859-1 -*-
+# -*- encoding: utf-8 -*-
 #
 #
-# Copyright (C) 2002-2004 Jrg Lehmann <joergl@users.sourceforge.net>
-# Copyright (C) 2003-2007 Michael Schindler <m-schindler@users.sourceforge.net>
-# Copyright (C) 2002-2006 Andr Wobst <wobsta@users.sourceforge.net>
+# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
+# Copyright (C) 2003-2011 Michael Schindler <m-schindler@users.sourceforge.net>
+# Copyright (C) 2002-2011 André Wobst <wobsta@users.sourceforge.net>
 #
 # This file is part of PyX (http://pyx.sourceforge.net/).
 #
@@ -21,10 +21,21 @@
 # along with PyX; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 
-import glob, os, threading, Queue, re, tempfile, atexit, time, warnings
-import config, siteconfig, unit, box, canvas, trafo, version, attr, style, dvifile
+import errno, glob, os, threading, Queue, re, tempfile, atexit, time, warnings
+import config, unit, box, canvas, trafo, version, attr, style, filelocator
+from pyx.dvi import dvifile
 import bbox as bboxmodule
 
+try:
+    import subprocess
+except ImportError:
+    have_subprocess = False
+else:
+    have_subprocess = True
+
+class PyXTeXWarning(UserWarning): pass
+warnings.filterwarnings('always', category=PyXTeXWarning)
+
 ###############################################################################
 # texmessages
 # - please don't get confused:
@@ -100,12 +111,6 @@ class _texmessagestart(texmessage):
             raise TexResultError("TeX startup failed", texrunner)
         texrunner.texmessageparsed = texrunner.texmessageparsed[m.end():]
 
-        # check for filename to be processed
-        try:
-            texrunner.texmessageparsed = texrunner.texmessageparsed.split("%s.tex" % texrunner.texfilename, 1)[1]
-        except (IndexError, ValueError):
-            raise TexResultError("TeX running startup file failed", texrunner)
-
         # check for \raiseerror -- just to be sure that communication works
         try:
             texrunner.texmessageparsed = texrunner.texmessageparsed.split("*! Undefined control sequence.\n<*> \\raiseerror\n               %\n", 1)[1]
@@ -182,23 +187,17 @@ class _texmessageend(texmessage):
 
     __implements__ = _Itexmessage
 
+    auxPattern = re.compile(r"\(([^()]+\.aux|\"[^\"]+\.aux\")\)")
+
     def check(self, texrunner):
-        try:
-            s1, s2 = texrunner.texmessageparsed.split("(%s.aux)" % texrunner.texfilename, 1)
-            texrunner.texmessageparsed = s1 + s2
-        except (IndexError, ValueError):
-            try:
-                s1, s2 = texrunner.texmessageparsed.split("(%s%s%s.aux)" % (os.curdir,
-                                                                        os.sep,
-                                                                        texrunner.texfilename), 1)
-                texrunner.texmessageparsed = s1 + s2
-            except (IndexError, ValueError):
-                pass
+        m = self.auxPattern.search(texrunner.texmessageparsed)
+        if m:
+            texrunner.texmessageparsed = (texrunner.texmessageparsed[:m.start()] + texrunner.texmessageparsed[m.end():]).strip()
 
         # check for "(see the transcript file for additional information)"
         try:
             s1, s2 = texrunner.texmessageparsed.split("(see the transcript file for additional information)", 1)
-            texrunner.texmessageparsed = s1 + s2
+            texrunner.texmessageparsed = (s1 + s2).strip()
         except (IndexError, ValueError):
             pass
 
@@ -250,35 +249,47 @@ class _texmessageload(texmessage):
 
     __implements__ = _Itexmessage
 
-    pattern = re.compile(r"\([\"]?(?P<filename>(?:(?<!\")[^()\s\n]+(?!\"))|[^()\"\n]+)[\"]?(?P<additional>[^()]*)\)")
+    pattern = re.compile(r"\([\"]?(?P<filename>(?:(?<!\")[^()\s\n]+(?!\"))|[^\"\n]+)[\"]?(?P<additional>[^()]*)\)")
 
-    def baselevels(self, s, maxlevel=1, brackets="()"):
+    def baselevels(self, s, maxlevel=1, brackets="()", quotes='""'):
         """strip parts of a string above a given bracket level
         - return a modified (some parts might be removed) version of the string s
           where all parts inside brackets with level higher than maxlevel are
           removed
         - if brackets do not match (number of left and right brackets is wrong
           or at some points there were more right brackets than left brackets)
-          just return the unmodified string"""
+          just return the unmodified string
+        - a quoted string immediately followed after a bracket is left untouched
+          even if it contains quotes itself"""
         level = 0
         highestlevel = 0
+        inquote = 0
         res = ""
-        for c in s:
-            if c == brackets[0]:
-                level += 1
-                if level > highestlevel:
-                    highestlevel = level
-            if level <= maxlevel:
+        for i, c in enumerate(s):
+            if quotes and level <= maxlevel:
+                if not inquote and c == quotes[0] and i and s[i-1] == brackets[0]:
+                    inquote = 1
+                elif inquote and c == quotes[1]:
+                    inquote = 0
+            if inquote:
                 res += c
-            if c == brackets[1]:
-                level -= 1
+            else:
+                if c == brackets[0]:
+                    level += 1
+                    if level > highestlevel:
+                        highestlevel = level
+                if level <= maxlevel:
+                    res += c
+                if c == brackets[1]:
+                    level -= 1
         if level == 0 and highestlevel > 0:
             return res
 
     def check(self, texrunner):
-        lowestbracketlevel = self.baselevels(texrunner.texmessageparsed)
-        if lowestbracketlevel is not None:
-            m = self.pattern.search(lowestbracketlevel)
+        search = self.baselevels(texrunner.texmessageparsed)
+        res = []
+        if search is not None:
+            m = self.pattern.search(search)
             while m:
                 filename = m.group("filename").replace("\n", "")
                 try:
@@ -287,20 +298,23 @@ class _texmessageload(texmessage):
                     additional = ""
                 if (os.access(filename, os.R_OK) or
                     len(additional) and additional[0] == "\n" and os.access(filename+additional.split()[0], os.R_OK)):
-                    lowestbracketlevel = lowestbracketlevel[:m.start()] + lowestbracketlevel[m.end():]
+                    res.append(search[:m.start()])
                 else:
-                    break
-                m = self.pattern.search(lowestbracketlevel)
+                    res.append(search[:m.end()])
+                search = search[m.end():]
+                m = self.pattern.search(search)
             else:
-                texrunner.texmessageparsed = lowestbracketlevel
+                res.append(search)
+                texrunner.texmessageparsed = "".join(res)
 
 
 class _texmessageloaddef(_texmessageload):
     """validates the inclusion of font description files (fd-files)
     - works like _texmessageload
-    - filename must end with .def or .fd and no further text is allowed"""
+    - filename must end with .def or .fd
+    - further text is allowed"""
 
-    pattern = re.compile(r"\((?P<filename>[^)]+(\.fd|\.def))\)")
+    pattern = re.compile(r"\([\"]?(?P<filename>(?:(?:(?<!\")[^\(\)\s\n\"]+)|(?:(?<=\")[^\(\)\"]+))(\.fd|\.def))[\"]?[\s\n]*(?P<additional>[\(]?[^\(\)]*[\)]?)[\s\n]*\)")
 
     def baselevels(self, s, **kwargs):
         return s
@@ -375,6 +389,8 @@ class texmessagepattern(texmessage):
 texmessage.fontwarning = texmessagepattern(re.compile(r"^LaTeX Font Warning: .*$(\n^\(Font\).*$)*", re.MULTILINE), "ignoring font warning")
 texmessage.boxwarning = texmessagepattern(re.compile(r"^(Overfull|Underfull) \\[hv]box.*$(\n^..*$)*\n^$\n", re.MULTILINE), "ignoring overfull/underfull box warning")
 texmessage.rerunwarning = texmessagepattern(re.compile(r"^(LaTeX Warning: Label\(s\) may have changed\. Rerun to get cross-references right\s*\.)$", re.MULTILINE), "ignoring rerun warning")
+texmessage.packagewarning = texmessagepattern(re.compile(r"^package\s+(?P<packagename>\S+)\s+warning\s*:[^\n]+(?:\n\(?(?P=packagename)\)?[^\n]*)*", re.MULTILINE | re.IGNORECASE), "ignoring generic package warning")
+texmessage.nobblwarning = texmessagepattern(re.compile(r"^[\s\*]*(No file .*\.bbl.)\s*", re.MULTILINE), "ignoring no-bbl warning")
 
 
 
@@ -624,12 +640,13 @@ class _readpipe(threading.Thread):
     def run(self):
         """thread routine"""
         def _read():
+            # catch interupted system call errors while reading
             while 1:
                 try:
                     return self.pipe.readline()
                 except IOError, e:
-                    if e.errno not in (4, ): # interrupted
-                        raise
+                    if e.errno != errno.EINTR:
+                         raise
         read = _read() # read, what comes in
         try:
             self.expect = self.expectqueue.get_nowait() # read, what should be expected
@@ -736,7 +753,7 @@ def _cleantmp(texrunner):
             os.rename(texrunner.texfilename + usefile[extpos:], usefile)
         except OSError:
             pass
-    for file in glob.glob("%s.*" % texrunner.texfilename):
+    for file in glob.glob("%s.*" % texrunner.texfilename) + ["%sNotes.bib" % texrunner.texfilename]:
         try:
             os.unlink(file)
         except OSError:
@@ -768,17 +785,16 @@ class texrunner:
     defaulttexmessagesstart = [texmessage.start]
     defaulttexmessagesdocclass = [texmessage.load]
     defaulttexmessagesbegindoc = [texmessage.load, texmessage.noaux]
-    defaulttexmessagesend = [texmessage.end, texmessage.fontwarning, texmessage.rerunwarning]
+    defaulttexmessagesend = [texmessage.end, texmessage.fontwarning, texmessage.rerunwarning, texmessage.nobblwarning]
     defaulttexmessagesdefaultpreamble = [texmessage.load]
     defaulttexmessagesdefaultrun = [texmessage.loaddef, texmessage.graphicsload,
-                                    texmessage.fontwarning, texmessage.boxwarning]
+                                    texmessage.fontwarning, texmessage.boxwarning, texmessage.packagewarning]
 
     def __init__(self, mode="tex",
                        lfs="10pt",
                        docclass="article",
                        docopt=None,
                        usefiles=[],
-                       fontmaps=config.get("text", "fontmaps", "psfonts.map"),
                        waitfortex=config.getint("text", "waitfortex", 60),
                        showwaitfortex=config.getint("text", "showwaitfortex", 5),
                        texipc=config.getboolean("text", "texipc", 0),
@@ -800,7 +816,6 @@ class texrunner:
         self.docclass = docclass
         self.docopt = docopt
         self.usefiles = usefiles[:]
-        self.fontmaps = fontmaps
         self.waitfortex = waitfortex
         self.showwaitfortex = showwaitfortex
         self.texipc = texipc
@@ -853,9 +868,9 @@ class texrunner:
                 hasevent = event.isSet()
                 if not hasevent:
                     if waited < self.waitfortex:
-                        warnings.warn("still waiting for %s after %i (of %i) seconds..." % (self.mode, waited, self.waitfortex))
+                        warnings.warn("still waiting for %s after %i (of %i) seconds..." % (self.mode, waited, self.waitfortex), PyXTeXWarning)
                     else:
-                        warnings.warn("the timeout of %i seconds expired and %s did not respond." % (waited, self.mode))
+                        warnings.warn("the timeout of %i seconds expired and %s did not respond." % (waited, self.mode), PyXTeXWarning)
             return hasevent
         else:
             event.wait(self.waitfortex)
@@ -890,11 +905,16 @@ class texrunner:
                 ipcflag = " --ipc"
             else:
                 ipcflag = ""
-            try:
-                self.texinput, self.texoutput = os.popen4("%s%s %s" % (self.mode, ipcflag, self.texfilename), "t", 0)
-            except ValueError:
-                # XXX: workaround for MS Windows (bufsize = 0 makes trouble!?)
-                self.texinput, self.texoutput = os.popen4("%s%s %s" % (self.mode, ipcflag, self.texfilename), "t")
+            if have_subprocess:
+                p = subprocess.Popen("%s%s %s" % (self.mode, ipcflag, self.texfilename), shell=True, bufsize=0,
+                                     stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, close_fds=True)
+                self.texinput, self.texoutput = p.stdin, p.stdout
+            else:
+                try:
+                    self.texinput, self.texoutput = os.popen4("%s%s %s" % (self.mode, ipcflag, self.texfilename), "t", 0)
+                except ValueError:
+                    # XXX: workaround for MS Windows (bufsize = 0 makes trouble!?)
+                    self.texinput, self.texoutput = os.popen4("%s%s %s" % (self.mode, ipcflag, self.texfilename), "t")
             atexit.register(_cleantmp, self)
             self.expectqueue = Queue.Queue(1)  # allow for a single entry only -> keeps the next InputMarker to be wait for
             self.gotevent = threading.Event()  # keeps the got inputmarker event
@@ -902,7 +922,6 @@ class texrunner:
             self.quitevent = threading.Event() # keeps for end of terminal event
             self.readoutput = _readpipe(self.texoutput, self.expectqueue, self.gotevent, self.gotqueue, self.quitevent)
             self.texruns = 1
-            self.fontmap = dvifile.readfontmap(self.fontmaps.split())
             oldpreamblemode = self.preamblemode
             self.preamblemode = 1
             self.readoutput.start()
@@ -934,54 +953,27 @@ class texrunner:
             os.remove("%s.tex" % self.texfilename)
             if self.mode == "tex":
                 if self.lfs:
-                    lfserror = None
-                    if len(self.lfs) > 4 and self.lfs[-4:] == ".lfs":
-                        lfsname = self.lfs
-                    else:
-                        lfsname = "%s.lfs" % self.lfs
-                    for fulllfsname in [lfsname,
-                                        os.path.join(siteconfig.lfsdir, lfsname)]:
-                        try:
-                            lfsfile = open(fulllfsname, "r")
-                            lfsdef = lfsfile.read()
-                            lfsfile.close()
-                            break
-                        except IOError:
-                            pass
-                    else:
-                        lfserror = "File '%s' is not available or not readable. " % lfsname
-                else:
-                    lfserror = ""
-                if lfserror is not None:
-                    allfiles = (glob.glob("*.lfs") +
-                                glob.glob(os.path.join(siteconfig.lfsdir, "*.lfs")))
-                    lfsnames = []
-                    for f in allfiles:
-                        try:
-                            open(f, "r").close()
-                            lfsnames.append(os.path.basename(f)[:-4])
-                        except IOError:
-                            pass
-                    lfsnames.sort()
-                    if len(lfsnames):
-                        raise IOError("%sAvailable LaTeX font size files (*.lfs): %s" % (lfserror, lfsnames))
-                    else:
-                        raise IOError("%sNo LaTeX font size files (*.lfs) available. Check your installation." % lfserror)
-                self.execute(lfsdef, [])
-                self.execute("\\normalsize%\n", [])
+                    if not self.lfs.endswith(".lfs"):
+                        self.lfs = "%s.lfs" % self.lfs
+                    lfsfile = filelocator.open(self.lfs, [], "r")
+                    lfsdef = lfsfile.read()
+                    lfsfile.close()
+                    self.execute(lfsdef, [])
+                    self.execute("\\normalsize%\n", [])
                 self.execute("\\newdimen\\linewidth\\newdimen\\textwidth%\n", [])
             elif self.mode == "latex":
                 if self.pyxgraphics:
-                    pyxdef = os.path.join(siteconfig.sharedir, "pyx.def")
-                    try:
-                        open(pyxdef, "r").close()
-                    except IOError:
-                        IOError("file 'pyx.def' is not available or not readable. Check your installation or turn off the pyxgraphics option.")
-                    pyxdef = os.path.abspath(pyxdef).replace(os.sep, "/")
+                    pyxdef = filelocator.open("pyx.def", [], "rb")
+                    pyxdef_filename = self.texfilename + ".pyxdef"
+                    pyxdef_file = open(pyxdef_filename, "wb")
+                    pyxdef_file.write(pyxdef.read())
+                    pyxdef.close()
+                    pyxdef_file.close()
+                    pyxdef_filename_tex = os.path.abspath(pyxdef_filename).replace(os.sep, "/")
                     self.execute("\\makeatletter%\n"
                                  "\\let\\saveProcessOptions=\\ProcessOptions%\n"
                                  "\\def\\ProcessOptions{%\n"
-                                 "\\def\\Gin@driver{" + pyxdef + "}%\n"
+                                 "\\def\\Gin@driver{" + pyxdef_filename_tex + "}%\n"
                                  "\\def\\c@lor@namefile{dvipsnam.def}%\n"
                                  "\\saveProcessOptions}%\n"
                                  "\\makeatother",
@@ -1051,10 +1043,10 @@ class texrunner:
         self.execute(None, self.defaulttexmessagesend + self.texmessagesend)
         dvifilename = "%s.dvi" % self.texfilename
         if not self.texipc:
-            self.dvifile = dvifile.dvifile(dvifilename, self.fontmap, debug=self.dvidebug)
+            self.dvifile = dvifile.DVIfile(dvifilename, debug=self.dvidebug)
             page = 1
             for box in self.needdvitextboxes:
-                box.setdvicanvas(self.dvifile.readpage([ord("P"), ord("y"), ord("X"), page, 0, 0, 0, 0, 0, 0]))
+                box.setdvicanvas(self.dvifile.readpage([ord("P"), ord("y"), ord("X"), page, 0, 0, 0, 0, 0, 0], fontmap=box.fontmap))
                 page += 1
         if not ignoretail and self.dvifile.readpage(None) is not None:
             raise RuntimeError("end of dvifile expected")
@@ -1086,7 +1078,6 @@ class texrunner:
                   docclass=_unset,
                   docopt=_unset,
                   usefiles=_unset,
-                  fontmaps=_unset,
                   waitfortex=_unset,
                   showwaitfortex=_unset,
                   texipc=_unset,
@@ -1119,8 +1110,6 @@ class texrunner:
             self.docopt = docopt
         if usefiles is not _unset:
             self.usefiles = usefiles
-        if fontmaps is not _unset:
-            self.fontmaps = fontmaps
         if waitfortex is not _unset:
             self.waitfortex = waitfortex
         if showwaitfortex is not _unset:
@@ -1174,7 +1163,7 @@ class texrunner:
 
     PyXBoxPattern = re.compile(r"PyXBox:page=(?P<page>\d+),lt=(?P<lt>-?\d*((\d\.?)|(\.?\d))\d*)pt,rt=(?P<rt>-?\d*((\d\.?)|(\.?\d))\d*)pt,ht=(?P<ht>-?\d*((\d\.?)|(\.?\d))\d*)pt,dp=(?P<dp>-?\d*((\d\.?)|(\.?\d))\d*)pt:")
 
-    def text(self, x, y, expr, textattrs=[], texmessages=[]):
+    def text(self, x, y, expr, textattrs=[], texmessages=[], fontmap=None):
         """create text by passing expr to TeX/LaTeX
         - returns a textbox containing the result from running expr thru TeX/LaTeX
         - the box center is set to x, y
@@ -1204,12 +1193,16 @@ class texrunner:
             expr = textattrs[lentextattrs-1-i].apply(expr)
         try:
             self.execute(expr, self.defaulttexmessagesdefaultrun + self.texmessagesdefaultrun + texmessages)
-        except TexResultError:
-            self.finishdvi(ignoretail=1)
-            raise
+        except TexResultError, e:
+            warnings.warn("We try to finish the dvi due to an unhandled tex error", PyXTeXWarning)
+            try:
+                self.finishdvi(ignoretail=1)
+            except TexResultError:
+                pass
+            raise e
         if self.texipc:
             if first:
-                self.dvifile = dvifile.dvifile("%s.dvi" % self.texfilename, self.fontmap, debug=self.dvidebug)
+                self.dvifile = dvifile.DVIfile("%s.dvi" % self.texfilename, debug=self.dvidebug)
         match = self.PyXBoxPattern.search(self.texmessage)
         if not match or int(match.group("page")) != self.page:
             raise TexResultError("box extents not found", self)
@@ -1220,8 +1213,9 @@ class texrunner:
                                 #       this is quite different from what we do elsewhere!!!
                                 #       see https://sourceforge.net/mailarchive/forum.php?thread_id=9137692&forum_id=23700
         if self.texipc:
-            box.setdvicanvas(self.dvifile.readpage([ord("P"), ord("y"), ord("X"), self.page, 0, 0, 0, 0, 0, 0]))
+            box.setdvicanvas(self.dvifile.readpage([ord("P"), ord("y"), ord("X"), self.page, 0, 0, 0, 0, 0, 0], fontmap=fontmap))
         else:
+            box.fontmap = fontmap
             self.needdvitextboxes.append(box)
         return box
 
@@ -1270,7 +1264,7 @@ class texrunner:
                          "\\vfill\\supereject%%\n" % text, [texmessage.ignore])
             if self.texipc:
                 if self.dvifile is None:
-                    self.dvifile = dvifile.dvifile("%s.dvi" % self.texfilename, self.fontmap, debug=self.dvidebug)
+                    self.dvifile = dvifile.DVIfile("%s.dvi" % self.texfilename, debug=self.dvidebug)
             else:
                 raise RuntimeError("textboxes currently needs texipc")
             lastparnos = parnos
